﻿/*
 *  TI Unresolved References 0.7
 *  Copyright (C) 2011  Ralph Schenn rschenn@t-online.de
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************
 *
 * 19.10.11 optional auch Funktionen ohne Referenzen auflisten
 *          gemeinsamer EventHandler fuer alle Checkboxes
 *          Accelerators Ctrl-O und Ctrl-S
 *          im zuletzt ausgewaehlten Pfad bleiben
 * 18.10.11 neue Methode Scan
 * 17.10.11 ConverAll statt ConvertName
 *          Loeschen Verzeichnis kann fehlschlagen
 *          Pruefung ob \ am Ende der Funktion steht vorgezogen
 * 08.10.11 ShowDialog statt Show
 * 07.10.11 griechische Zeichen im TreeView ausgeschrieben anzeigen
 * 05.10.11 Core.Error verwenden
 *          Begin-/EndUpdate beim expandieren/kollabieren/loeschen des TreeView
 *          erst Select dann Scan als default-Button
 *          wenn missing zu Aufrufer auf true gesetzt wurde darf dies erst beim
 *          naechsten Aufrufer wieder auf false gesetzt werden
 * 04.07.11 Hinweis Mehrfachaufruf nur 1x dranhaengen
 *          neue Methode Folder
 *          Inhalt des Ordners unter gemeinsamem Knoten darstellen
 * 30.06.11 nach Anklicken Funktion Inhalt der Funktion anzeigen
 *          neue Methode SubNode
 *          Verzeichnis vor Scan und am Programmende loeschen
 * 21.06.11 fehlende Referenz unter derselben Funktion nur 1x listen/zaehlen
 *          Voyage 200 group file extensions ergaenzt
 *          (n) anhaengen wenn fehlende Referenz in derselben Funktion 
 *          mehrfach aufgerufen wird
 * 19.06.11 falsches Dateiformat behandeln
 * 17.06.11 File.Exists statt Directory.Exists
 *          OpenFileDialog statt FolderBrowserDialog
 *          Archiv entpacken
 *          Verzeichnis danach loeschen
 * 16.06.11 Progressionsbalken anzeigen
 * 06.06.11 nach manueller Eingabe Pfad Button Scan nur aktivieren wenn das 
 *          Verzeichnis existiert
 *          Formular auch dann schliessen wenn tbPath erfolglos validiert wurde
 *          Init() ueberarbeitet
 *          bei leerem Verzeichnis Mauszeiger zuruecksetzen
 *          Resize zulassen
 * 05.06.11 neue Methode Init
 *          Assemblerdateien an auf z endender Extension erkennen
 *          Meldung wenn gewaehltes Verzeichnis leer ist
 *          Anzahl gescannter Dateien anzeigen
 * 01.06.11 TreeView nach Aenderung Checkboxes leeren
 *          Button Scan (de)aktivieren
 * 29.05.11 this verwenden
 * 25.05.11 FxCop
 * 24.05.11 GetFileNameWithoutExtension verwenden
 *          wenn keine fehlenden Referenzen angezeigt werden sollen auch              
 *          uebergeordnete Funktionen nicht auflisten
 *          Sanduhr anzeigen
 *          bei Select TreeView loeschen
 * 23.05.11 Resize deaktiviert
 * 19.05.11 Anzahl fehlender Referenzen anzeigen
 * 19.05.11 Ersterstellung
 * 
 */

using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Windows.Forms;

namespace UnresRef
{
    public partial class MainForm : Form
    {
        #region constants

        // Trennzeichen extension
        private const char dot = '.';

        #endregion

        #region instance variables

        // Kurzbefehle
        private enum accel
        {
            undef,
            path,
            scan
        };

        // Mapping Kurzbefehle zu Tastenkombination
        Hashtable accelHash = null;

        // Liste mit Dateinamen
        private List<string> files = null;

        // Anzahl gescannter Dateien
        private int scanned = 0;

        // Anzahl offener Referenzen
        private int unres = 0;

        // TreeView aufgeklappt
        private bool expanded = false;

        // zuletzt ausgewaehlter Ordner
        private string selPath = null;

        // Pfad zum entpackten Archiv
        private string path = null;

        #endregion

        /// <summary>
        /// Konstruktor Formular
        /// </summary>
        /// <param name="path">Pfad TI group file</param>
        public MainForm(string path)
        {
            InitializeComponent();

            // den Kurzbefehlen Tasten zuweisen
            MapAccelerators();

            // Name und Version in Titelleiste anzeigen
            this.Text = Application.ProductName + " " + 
                Application.ProductVersion;

            // wenn Pfad gueltig ist
            if (path != null)
            {
                // Pfad TI group file in TextBox anzeigen
                this.tbPath.Text = path;
            }

            // Formular initialisieren
            Init();

            // default-Button bei Enter
            this.AcceptButton = bSelect;
        }

        /// <summary>
        /// Behandlung Kurzbefehle
        /// </summary>
        /// <param name="msg"></param>
        /// <param name="keyData"></param>
        /// <returns>Button wurde behandelt</returns>
        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            // Check this key...
            bool bHandled = false;

            // Look up value
            accel accel = accel.undef;

            if (this.accelHash.ContainsKey(new AcceleratorKey(keyData)))
            {
                accel = (accel)this.accelHash[keyData];

                switch (accel)
                {
                    case accel.path:
                        bPath_Click(null, null);
                        bHandled = true;
                        break;

                    case accel.scan:
                        bScan_Click(null, null);
                        bHandled = true;
                        break;

                    case accel.undef:
                    default:
                        break;
                }
            }

            return bHandled;
        }

        #region event handlers

        /// <summary>
        /// Pfad auswaehlen
        /// </summary>
        private void bPath_Click(object sender, System.EventArgs e)
        {           
            // Pfad abfragen
            tbPath.Text = GetPath();

            // Formular initialisieren
            Init();

            // default-Button bei Enter
            this.AcceptButton = bScan;
        }

        /// <summary>
        /// Funktionen nach fehlenden Referenzen durchsuchen
        /// </summary>
        private void bScan_Click(object sender, System.EventArgs e)
        {

            // Pfad Verzeichnis in das TI group file entpackt wird
            this.path = Path.Combine(Path.GetDirectoryName(tbPath.Text),
                Path.GetFileNameWithoutExtension(tbPath.Text));

            // Sanduhr
            this.Cursor = Cursors.WaitCursor;

            // Formular initialisieren
            Init();

            // wenn Verzeichnis existiert
            if (Directory.Exists(this.path))
            {
                try
                {
                    // Verzeichnis loeschen
                    Directory.Delete(this.path, true);
                }
                catch (Exception ex)
                {
                    Core.Error(ex);
                    return;
                }
            }
            
            // TI group file erfolgreich in temp. Verzeichnis entpackt
            if (Core.Unzip(tbPath.Text))
            {
                // fehlende Referenzen suchen und Ergebnis anzeigen
                Scan(tvResults, prbProgress, lblState, this.path);
            }
            // die Datei ist keine zip-Datei
            else
            {
                Core.Error("The file you have selected is not a TI group file");
            }
             
            // normaler Mauszeiger
            this.Cursor = Cursors.Default;
        }

        /// <summary>
        /// fehlende Referenzen suchen und anzeigen
        /// </summary>
        /// <param name="view">Baum mit Ergebnissen</param>
        /// <param name="bar">Fortschrittsbalken</param>
        /// <param name="lbl">Statuszeile</param>
        /// <param name="path">Pfad zum Ordner mit entpacktem Archiv</param>
        private void Scan(TreeView view, ProgressBar bar, Label lbl, string path)
        {
            // Klasse zum Zugriff auf Dateien
            Core core = new Core(path);

            // zuletzt verarbeiteter Ordner
            string previousFolder = null;
            
            // laufende Updates vermeiden
            view.BeginUpdate();

            // Liste der Funktionen im Ordner ermitteln
            this.files = core.GetFiles();

            // mindestens eine Funktion im Verzeichnis gefunden
            if (this.files != null)
            {
                try
                {
                    // Fortschrittsbalken initialisieren
                    bar.Maximum = this.files.Count;
                    bar.Step = 1;
                    bar.Visible = true;

                    TreeNode folderNode = null;

                    // Iteration ueber alle Funktionen im Ordner
                    foreach (string file in this.files)
                    {
                        string folderName = GetFolderName(file);

                        // Gruppenwechsel Ordner
                        if (folderName != previousFolder)
                        {
                            // Ordner-Knoten angelegt
                            if (folderNode != null)
                            {
                                // wenn Referenzen fehlen und das nicht 
                                // angezeigt werden soll Knoten entfernen
                                RemoveNode(folderNode.ForeColor == Color.Red, folderNode);
                            }

                            // Ordner zum TreeView hinzufuegen
                            folderNode = tvResults.Nodes.Add(folderName);
                        }

                        // aufrufende Datei enthaelt fehlende Referenzen
                        if (CallerContUnresRef(folderNode, file))
                        {
                            // Ordner rot darstellen
                            folderNode.ForeColor = Color.Red;
                        }

                        // letzten Ordner speichern
                        previousFolder = GetFolderName(file);
                    }

                    // wenn Referenzen fehlen und das nicht 
                    // angezeigt werden soll Knoten entfernen
                    // (hier noch mal aufrufen damit auch der letzte Ordner 
                    // geprueft wird) 
                    RemoveNode(folderNode.ForeColor == Color.Red, folderNode);

                    // Fortschrittsbalken verbergen
                    bar.Visible = false;
                    bar.Value = 0;
                }
                catch (Exception ex)
                {
                    Core.Error(ex);
                    return;
                }

                view.EndUpdate();

                // Anzahl offener Referenzen/gescannter Dateien anzeigen
                lbl.Text = String.Format("{0} unresolved references found ({1} files scanned)",
                    this.unres,
                    this.scanned);
            }
            // Verzeichnis ist leer
            else
            {
                lbl.Text = "The TI group file is empty";
            }

            core.Dispose();
        }

        /// <summary>
        /// Status beliebiger Ckeckbox geaendert
        /// </summary>
        private void CB_CheckedChanged(object sender, EventArgs e)
        {
            // Formular initialisieren
            Init();
        }
        
        /// <summary>
        /// es wurde manuell ein Pfad eingegeben
        /// </summary>
        private void tbPath_Validating(object sender, System.ComponentModel.CancelEventArgs e)
        {
            // die Datei existiert
            if (File.Exists(tbPath.Text))
            {
                e.Cancel = false;
            }
            // die Datei existiert nicht
            else
            {
                Core.Error("The path you have entered does not exist");

                e.Cancel = true;
            }

            // Formular initialisieren
            Init();
        }

        /// <summary>
        /// Feld Pfad betreten
        /// </summary>
        private void tbPath_Enter(object sender, EventArgs e)
        {
            // ??? funktioniert nicht immer
            tbPath.SelectAll();
        }

        /// <summary>
        /// Formular wird geschlossen
        /// </summary>
        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            // Formular auch dann schliessen wenn in tbPath ein ungueltiger Pfad 
            // eingetragen ist
            e.Cancel = false;

            // wenn Verzeichnis existiert
            if (Directory.Exists(this.path))
            {
                try
                {
                    // Verzeichnis loeschen
                    Directory.Delete(this.path, true);
                }
                catch (Exception ex)
                {
                    Core.Error(ex);
                }
            }
        }

        /// <summary>
        /// Pfad hat sich geaendert
        /// </summary>
        private void tbPath_TextChanged(object sender, EventArgs e)
        {
            // Formular initialisieren
            Init();

            try
            {
                // versuchen Eventhandler zu trennen (nur registriert wenn 
                // Pfad bereits geaendert wurde)
                this.tbPath.Validating -= new System.ComponentModel.CancelEventHandler(this.tbPath_Validating);
            }
            catch
            {
                // kein Fehler (Pfad wurde noch nie geaendert)
            }
            finally
            {
                // Validator Pfad registrieren
                this.tbPath.Validating += new System.ComponentModel.CancelEventHandler(this.tbPath_Validating);
            }
        }

        /// <summary>
        /// angeklickte Funktion anzeigen
        /// </summary>
        private void tvResults_AfterSelect(object sender, TreeViewEventArgs e)
        {
            // aufrufende Funktion angeklickt (Dateiname mit Extension)
            if (e.Node.Text.Split(dot).Length == 3)
            {
                // neues Formular zur Anzeige der Funktion
                Content c = new Content();

                // Quellcode uebergeben
                c.SetText(e.Node.Text, Core.GetFunction(Path.Combine(this.path, e.Node.Text)));

                // Formular anzeigen
                c.ShowDialog();
            }
        }

        /// <summary>
        /// TreeView auf-/zuklappen
        /// </summary>
        private void tvResults_DoubleClick(object sender, EventArgs e)
        {
            tvResults.BeginUpdate();

            // TreeView bereits aufgeklappt
            if (this.expanded)
            {
                tvResults.CollapseAll();
                this.expanded = false;
            }
            // TreeView geschlossen
            else
            {
                tvResults.ExpandAll();
                this.expanded = true;
            }

            tvResults.EndUpdate();
        }
        
        #endregion

        #region private methods

        /// <summary>
        /// Ordnername aus Dateiname extrahieren
        /// </summary>
        /// <param name="fileName">Dateiname</param>
        /// <returns>Ordnername</returns>
        private string GetFolderName(string fileName)
        {
            return fileName.Substring(0, fileName.IndexOf(dot));
        }

        /// <summary>
        /// Ordner verarbeiten
        /// </summary>
        /// <param name="folderNode">Knoten des Ordners</param>
        /// <param name="file">Name der aufrufenden Datei</param>
        /// <returns>Referenz fehlt</returns>
        private bool CallerContUnresRef(TreeNode folderNode, string file)
        {
            // Referenzen in Funktion
            List<string> refs = null;

            // Funktion hat keine offenen Referenzen
            bool missing = false;

            // Knoten des Aufrufers
            TreeNode callerNode = null;
            
            // ASM-Dateien nicht durchsuchen
            if (!Path.GetExtension(file).EndsWith("z"))
            {
                // Anzahl gescannter Dateien erhoehen
                this.scanned++;

                // die Referenz fehlt nicht
                missing = false;

                // Knoten des Aufrufers suchen (Ordner)
                callerNode = GetSubnode(folderNode, file);

                // aufrufende Funktion noch nicht hinzugefuegt
                if (callerNode == null)
                {
                    // aufrufende Funktion hinzufuegen
                    callerNode = folderNode.Nodes.Add(file);
                }

                // Referenzen in aufrufender Funktion suchen
                refs = Core.GetReferences(Path.Combine(this.path, file));

                // mindestens eine Referenz in aufrufender Funktion gefunden
                if (refs != null)
                {
                    // Referenzen zur aufrufenden Funktion hinzufuegen
                    foreach (string r in refs)
                    {
                        // wenn bei einem Aurufer eine Referenz gefehlt hat
                        // darf missing bei folgenden Referenzen in demselben 
                        // Aufrufer nicht wieder auf false gesetzt werden wenn
                        // die folgenden Referenzen vorhanden sind
                        if (ReferenceMissing(callerNode, r))
                        {
                            missing = true;
                        }
                    }
                }

                // Funktionen ohne Referenzen nicht anzeigen
                if (!cbNoRefs.Checked)
                {
                    RemoveNode(missing, callerNode);
                }
            }

            // Progressionsbalken verlaengern
            prbProgress.PerformStep();

            return missing;
        }

        /// <summary>
        /// Referenz pruefen und Ergebnis anzeigen
        /// </summary>
        /// <param name="callerNode">Name der aufrufenden Funktion</param>
        /// <param name="reference">Name der aufgerufenen Funktion</param>
        /// <returns>Referenz fehlt</returns>
        private bool ReferenceMissing(TreeNode callerNode, string reference)
        {
            // Referenz fehlt
            bool missing = false;

            // Trennzeichen Quellcode in Trennzeichen Datei umwandeln (. statt \)
            reference = reference.Replace(Core.DivFoldFunc, dot);

            // pruefen ob die aufgerufene Datei existiert
            bool found = FileExits(reference);

            // Referenz soll angezeigt werden
            if ((cbMissing.Checked && !found) || cbFound.Checked)
            {
                // bereits eingefuegte fehlende Referenz suchen
                TreeNode child = GetSubnode(callerNode, reference);

                // fehlende Referenz nicht gefunden
                if (child == null)
                {
                    // aufgerufene Funktion zur aufrufenden Funktion hinzufuegen
                    TreeNode subNode = callerNode.Nodes.Add(Core.ConvertAll(Core.Greek, reference));

                    // aufgerufene Funktion existiert nicht
                    if (!found)
                    {
                        missing = true;

                        // aufrufende Funktion mit fehlenden Referenzen kennzeichnen
                        callerNode.ForeColor = Color.Red;

                        // fehlende aufgerufene Funktion kennzeichnen
                        subNode.ForeColor = Color.Red;

                        // Anzahl fehlender Referenzen erhoehen
                        this.unres++;
                    }
                }
                // fehlende Referenz bereits hinzugefuegt
                else
                {
                    // Kennzeichnung Mehrfachaufruf
                    const string multiple = " (n)";

                    // noch kein Hinweis auf Mehrfachaufruf angehaengt
                    if (!child.Text.EndsWith(multiple))
                    {
                        // Hinweis Mehrfachaufruf anhaengen
                        child.Text = child.Text + multiple;                        
                    }
                }
            }

            return missing;
        }

        /// <summary>
        /// Dialog Auswahl Pfad
        /// </summary>
        /// <returns>Pfad zur TI group file</returns>
        private string GetPath()
        {
            // Pfad zur TI group file
            string path = null;

            // neuer Dialog zur Auswahl der TI group file
            OpenFileDialog dialog = new OpenFileDialog();

            // noch keine Datei ausgewaehlt
            if (this.selPath == null)
            {
                // nach Start des Programms Eigene Dateien
                dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            }
            else
            {
                // nach erster Auwahl im Verzeichnis bleiben
                dialog.InitialDirectory = this.selPath;
            }

            // Dialogtitel
            dialog.Title = "Select archive with functions";

            // nur Dateitypen auflisten
            dialog.Filter = "TI Group|*.tig|Voyage 200 Group|*.vig|All files|*.*";

            // es wurde eine Datei ausgewaehlt
            if (dialog.ShowDialog(this) == DialogResult.OK)
            {
                // Pfad der Datei
                path = dialog.FileName;
            }

            this.selPath = Path.GetDirectoryName(path);
            return path;
        }

        /// <summary>
        /// pruefen ob die referenzierte Datei existiert
        /// </summary>
        /// <param name="name">Name der Datei</param>
        /// <returns>Datei existiert</returns>
        private bool FileExits(string name)
        {
            // die Datei existiert
            bool exists = false;

            // Parameter ist gueltig
            if (!String.IsNullOrEmpty(name))
            {
                // Iteration ueber alle Dateien im Verzeichnis
                foreach (string file in this.files)
                {
                    // die referenzierte Datei existiert
                    if (file.StartsWith(Core.ConvertAll(Core.Greek, name)))
                    {
                        exists = true;
                        break;
                    }
                }
            }

            return exists;
        }

        /// <summary>
        /// Formular initialisieren
        /// </summary>
        private void Init()
        {
            // Anzahl gescannter Dateien initialisieren
            this.scanned = 0;

            // Anzahl fehlender Dateien initialisieren
            this.unres = 0;

            // Statuszeile initialisieren
            lblState.Text = String.Empty;

            // TreeView leeren
            tvResults.BeginUpdate();
            tvResults.Nodes.Clear();
            tvResults.EndUpdate();

            // Button Scan (de)aktivieren
            bScan.Enabled = File.Exists(tbPath.Text) && 
                (this.cbFound.Checked || this.cbMissing.Checked) ? true : false;
        }

        /// <summary>
        /// Kind unter Elternknoten suchen
        /// </summary>
        /// <param name="parent">Eltern-Knoten</param>
        /// <param name="child">Kind-Knoten</param>
        /// <returns>Kind</returns>
        private TreeNode GetSubnode(TreeNode parent, string child)
        {
            // gefundener Kind-Knoten
            TreeNode subNode = null;

            // Iteration ueber alle Kind-Knoten
            foreach (TreeNode node in parent.Nodes)
            {
                // Kind-Knoten gefunden
                if (node.Text.StartsWith(child))
                {
                    subNode = node;
                    break;
                }
            }

            return subNode;
        }

        /// <summary>
        /// wenn Referenzen fehlen und das nicht 
        /// angezeigt werden soll Knoten entfernen
        /// </summary>
        /// <param name="missing">es fehlen Referenzen</param>
        /// <param name="node">zu entfernder Knoten</param>
        private void RemoveNode(bool missing, TreeNode node)
        {
            // es fehlen keine Referenzen und
            // Funktionen ohne fehlende Referenzen nicht anzeigen
            if ((!missing && !cbFound.Checked) ||
                // es fehlen Referenzen und
                // Funktionen mit fehlenden Referenzen nicht anzeigen
                (missing && !cbMissing.Checked) ||
                // keine untergeordneten Knoten vorhanden
                node.Nodes.Count == 0)
            {
                // aufrufende Funktion entfernen
                node.Remove();
            }
        }

        /// <summary>
        /// den Kurzbefehlen Tasten zuweisen
        /// </summary>
        private void MapAccelerators()
        {
            this.accelHash = new Hashtable();

            this.accelHash.Add(new AcceleratorKey(Keys.Control | Keys.O),
                          accel.path);
            this.accelHash.Add(new AcceleratorKey(Keys.Control | Keys.S),
                          accel.scan);
        }
        
        #endregion

    }
}
